/**
@file 
@date        11/03/2015
@author      Josef Hofer 
@copyright   CERN (EN-ICE-SCD) 

@brief Library contains functions used by the unLHCLaser component.

Usage: Public

PVSS manager usage: CTRL, UI

Constraints:
- PVSS version: 3.11
- operating system: Linux.
*/


#uses "LHCServices_Config"	 //load dll or so file

// Input commands
const int LASER_MANAGER_COMMAND_IN_UNKNOWN = 0;
const int LASER_MANAGER_COMMAND_IN_START   = 10;
const int LASER_MANAGER_COMMAND_IN_ABORT   = 99;
const int LASER_MANAGER_COMMAND_IN_STOP    = 100;

// Manager states
const int LASER_MANAGER_STATE_UNKNOWN   = 1000;
const int LASER_MANAGER_STATE_INIT  	= 1005; 
const int LASER_MANAGER_STATE_ACTIVE 	= 1010;	
const int LASER_MANAGER_STATE_PASSIVE	= 1040;	
const int LASER_MANAGER_STATE_STOPPED 	= 1050;	
const int LASER_MANAGER_STATE_SUSPENDED = 1080;	
const int LASER_MANAGER_STATE_FINISHED 	= 1090;

// Manager infos, warnings and errors
const int LASER_MANAGER_MSG_INFO_BASE   = 10000;
const int LASER_MANAGER_MSG_WARNING_BASE= 10100;
const int LASER_MANAGER_MSG_ERROR_BASE  = 10200;

const int LASER_CONFIGPATH=1;
const int LASER_TCP_REMOTEHOST=2;
const int LASER_TCP_DATAPATH=3;
const int LASER_SHOWSTATISTICS=4;
const int LASER_TCP_MAXPACKETSIZE=5;
const int LASER_BACKUP_PERIOD=6;
const int LASER_LOG_LEVEL=7;
const int LASER_NUM_PERIOD=8;
const int LASER_T_MODE=9;

//State Manager
const int Laser_Command=1;
const int Laser_StateOut=2;
const int Laser_Message=3;
const int Laser_PVSS=4;

//Catalogue name
const string g_sCat_Laser="unLaser";

//user allowed to modify LASER if Unicos Package not installed
const string g_Laser_userAdmin="root"; 

// default settings
private const string DEFAULT_LASER_BROKER_PORT      = "2580";
private const string DEFAULT_LASER_CONFIG_FILE      = "data/LaserData.xml";
private const string DEFAULT_LASER_BROKER_LIST      = "LASER-TEST:2580";
private const string DEFAULT_LASER_TOPIC_BASE       = "EN_ICE_TEST";
private const string DEFAULT_LASER_STATISTIC_MODE   = "200";
private const string DEFAULT_LASER_T_MODE_SETTING   = "FALSE";
private const string DEFAULT_LASER_MAX_PACKET_SIZE  = "1000";
private const string DEFAULT_LASER_BACKUP_PERIOD    = "600";
private const string DEFAULT_LASER_LOG_LEVEL        = "6";
private const string DEFAULT_LASER_HEARTBEAT_PERIOD = "10";


/** Get laser manager default settings
@param  dsExInfo     output, error information
@return dsDefaultSettings   the settings
*/
public dyn_string unLaser_getManagerDefaultSettings(dyn_string &dsExInfo)
{    
    dyn_string dsDefaultSettings;
   	string sSystemName = getSystemName();
	
   	if (dpExists(sSystemName + "Laser_Info.defaultSettings")) {
		      int iRes=dpGet(sSystemName + "Laser_Info.defaultSettings.Settings", dsDefaultSettings);
        if (iRes < 0) {
			         fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPSETGETFAILED"),"");	
        } else {            
            dyn_string dsHardcodedDefaultSettings = unLaser_getHardcodedDefaultSettings();                        
            if(dynlen(dsDefaultSettings) != dynlen(dsHardcodedDefaultSettings)) {
                dsDefaultSettings = dsHardcodedDefaultSettings;
                iRes=dpSetWait(sSystemName + "Laser_Info.defaultSettings.Settings", dsDefaultSettings);                
            }         
        }       		
	   } else {        		
		      int iRes=dpCreate("Laser_Info","_Un_Laser_Config");
		      if (iRes < 0) {
			         fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPCREATEFAILURE"),"");	
		      } else {	
            dsDefaultSettings = unLaser_getHardcodedDefaultSettings();																																																																				
			         iRes=dpSetWait(sSystemName + "Laser_Info.defaultSettings.Settings", dsDefaultSettings);
        }		
    }
    
    return dsDefaultSettings;
}


private dyn_string unLaser_getHardcodedDefaultSettings()
{
    dyn_string dsSettings;
   	dsSettings[LASER_CONFIGPATH] = DEFAULT_LASER_CONFIG_FILE;
   	dsSettings[LASER_TCP_REMOTEHOST] = DEFAULT_LASER_BROKER_LIST;
   	dsSettings[LASER_TCP_DATAPATH] = DEFAULT_LASER_TOPIC_BASE;
   	dsSettings[LASER_SHOWSTATISTICS] = DEFAULT_LASER_STATISTIC_MODE;
    dsSettings[LASER_T_MODE] = DEFAULT_LASER_T_MODE_SETTING;
   	dsSettings[LASER_TCP_MAXPACKETSIZE] = DEFAULT_LASER_MAX_PACKET_SIZE;
   	dsSettings[LASER_BACKUP_PERIOD] = DEFAULT_LASER_BACKUP_PERIOD;
   	dsSettings[LASER_LOG_LEVEL] = DEFAULT_LASER_LOG_LEVEL;
   	dsSettings[LASER_NUM_PERIOD] = DEFAULT_LASER_HEARTBEAT_PERIOD;
    return dsSettings;        
}


/** Set laser manager settings to default settings
@param sDP          input, Laser Manager DP 
@param dsExInfo     output, error information
*/
public void unLaser_setManagerDefaultSettings(string sDP, dyn_string &dsExInfo)
{
    dyn_string dsSettings = unLaser_getManagerDefaultSettings(dsExInfo);

    if (dynlen(dsExInfo)==0) {
	       if (dpExists(sDP + ".Settings")) {
          		int iRes=dpSetWait(
                sDP + ".Settings.man_ConfigPath", dsSettings[LASER_CONFIGPATH],
            				sDP + ".Settings.man_ShowStatistics", dsSettings[LASER_SHOWSTATISTICS],
            				sDP + ".Settings.man_BackupPeriod", dsSettings[LASER_BACKUP_PERIOD],
            				sDP + ".Settings.tcp_LogLevel", dsSettings[LASER_LOG_LEVEL],
            				sDP + ".Settings.tcp_BrokerName", dsSettings[LASER_TCP_REMOTEHOST],
            				sDP + ".Settings.tcp_TopicBase", dsSettings[LASER_TCP_DATAPATH],
            				sDP + ".Settings.T_mode", dsSettings[LASER_T_MODE],        
            				sDP + ".Settings.tcp_MaxPacketSize", dsSettings[LASER_TCP_MAXPACKETSIZE],
            				sDP + ".Settings.man_NumPeriod", dsSettings[LASER_NUM_PERIOD],                
            				sDP + ".State.st_CommandIn", LASER_MANAGER_COMMAND_IN_STOP,			
            				sDP + ".State.st_StateOut", LASER_MANAGER_STATE_UNKNOWN,
            				sDP + ".State.ev_MessageOut", LASER_MANAGER_MSG_INFO_BASE);		//Unknow message
    		  } else {
		          fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPDOESNOTEXIST"),"");
        }
    }
}


/** Get laser manager settings
@param  sDP         input, Laser Manager DP 
@param  dsExInfo    output, error information
@return dsSettings  the requested settings       
*/
public dyn_string unLaser_getManagerSettings(string sDP, dyn_string &dsExInfo)
{
    dyn_string dsSettings; 
    
    if (dpExists(sDP + ".Settings")) {
	       int iRes=dpGet(
            sDP + ".Settings.man_ConfigPath", dsSettings[LASER_CONFIGPATH],
        				sDP + ".Settings.man_ShowStatistics", dsSettings[LASER_SHOWSTATISTICS],
        				sDP + ".Settings.man_BackupPeriod", dsSettings[LASER_BACKUP_PERIOD],
        				sDP + ".Settings.tcp_LogLevel", dsSettings[LASER_LOG_LEVEL],
        				sDP + ".Settings.tcp_BrokerName", dsSettings[LASER_TCP_REMOTEHOST],
        				sDP + ".Settings.tcp_TopicBase", dsSettings[LASER_TCP_DATAPATH],
        				sDP + ".Settings.tcp_MaxPacketSize", dsSettings[LASER_TCP_MAXPACKETSIZE],
        				sDP + ".Settings.man_NumPeriod", dsSettings[LASER_NUM_PERIOD],
            sDP + ".Settings.T_mode", dsSettings[LASER_T_MODE]);

        if (iRes<0) {
		          fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPGETERROR"),"");
        }		
    } else {
	       fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPDOESNOTEXIST"),"");
    }	
    
    return dsSettings;
}

/** Check the specified server list and add default port number if it not exists.
@param  serverList       broker list to fix
@return fixedServerList  fixed broker list       
*/
public string unLaser_fixServerList(string serverList)
{
  string fixedServerList = "";
  dyn_string servers = strsplit(serverList, ",");
  
  for(int i=1; i<=dynlen(servers); i++)
  {
    dyn_string splitServerPort = strsplit(servers[i], ":");
    //Check if port is defined, if not add the default port
    if(dynlen(splitServerPort) < 2) {
      servers[i] = servers[i] + ":" + DEFAULT_LASER_BROKER_PORT;
    }
    //Append server to the list
    fixedServerList += servers[i];    
    
    //If we are not in the last Server of the list , add comma
    if(i != dynlen(servers)) {
       fixedServerList += ",";
    }
  }

  return fixedServerList;
}


//Function to  hide the full path of the config file and to convert it to "LastFolder\Filename" format.
public string unLaser_hideFilePath(string filename)
{
  string sPath = getPath(DATA_REL_PATH);
		
  if ((filename != "") && (strpos (filename, sPath)==0)) {
    int index=strpos (filename, "data");		
    sPath=substr(filename, 0, index);
    sPath=substr(filename, index, strlen(filename)-strlen(sPath));
  }
  return sPath;
}


// Extract manager number from manager name
public int unLaser_getManagerNumberFromName(string sManagerName)
{
    int iManagerNumber;
	   string sNum = substr(sManagerName, strpos(sManagerName,c_Laser_Pattern) + strlen(c_Laser_Pattern));
	   sscanf(sNum,"%d",iManagerNumber);    
    return iManagerNumber;    
}


//------------------------------------------------------------------------------------------------------------------------

/** Convert config file in old text format to xml format that can be read by the API-Manager

@param  	filePathName, string, in/out, full pathname of the config file
@return  true on success, false on error
*/
public bool unLaser_convertConfigFileToXml(string &filePathName)
{
  mapping data = unLaser_parse4XMLconversion(filePathName);           
  filePathName = unLaser_convertToXML(filePathName, data);   
  return true;  
}


/** Parse text config file and return the content

@param: 	sConfigPath, string, input, Path
@return:    mResult,     mapping, containing config file's content

@TODO: Re-implement this function in an understandable way and eliminate dependency on LHCServices_Config
*/
private mapping unLaser_parse4XMLconversion(string sConfigPath)
{
  mapping mResult;
  
  dyn_string Select; 
  dyn_string Type;  
  dyn_string Alias; 
  dyn_string Dpe; 
  dyn_string Name; 
  dyn_string Description;
  dyn_string Range; 
  dyn_string State; 
  dyn_string FF; 
  dyn_string FM; 
  dyn_string FC;   
  dyn_string Limit; 
  dyn_int Line; 
  bool bError;  
   
  string sSelect, sType, sDescription, sName, sDpe, sFC, sFM, sFF, sRange, sState, sAlias, sLimit;
  int i, len, iLen_Range,j,len2,iNbLine=0,iStart ;
  int c_DESCRIPTION=2, c_FF=3, c_FM=4, c_FC=5, c_STATE=6, c_Limit=7;
  float fMax, fIndex;
  dyn_string dsFileLines, dsSplitRange;
  dyn_dyn_string ddsFile;
  bool bDefined;  
  string ERROR_VALUE="";
  
    progressBar.visible(true);
  
    if (isFunctionDefined("ctrlConfRead"))
    {
        ctrlConfRead(sConfigPath,ddsFile);
        
		len=dynlen(ddsFile);
		fMax=len;
		for (i=1;i<=len;i++) //read each statement
		{				
			int l=1;
			dsFileLines=ddsFile[i];
			len2=dynlen(dsFileLines);
			fIndex=i;

			progressBar.progress(fIndex/fMax*100);
			if(patternMatch("*[*]*", dsFileLines[l]))
			{
				switch (dsFileLines[l]) 
                {
					case "[BOOLEAN]":
					case "[FLOAT]":
					case "[INTEGER]":
					case "[UNSIGNED]":
                        sType=dsFileLines[l];
                        bDefined=true; 
						break;
					default:
						bDefined=false;
				}	
						
				if(len2!=1) {
					dynAppend(Select,"F");
					dynAppend(Type,sType);
					dynAppend(Dpe,ERROR_VALUE);
					dynAppend(Alias,ERROR_VALUE);
					dynAppend(Line,iNbLine);		
					dynAppend(Description,ERROR_VALUE);
					dynAppend(Range,ERROR_VALUE);
					dynAppend(State,ERROR_VALUE);
					dynAppend(FF,ERROR_VALUE);
					dynAppend(FC,ERROR_VALUE);
					dynAppend(FM,ERROR_VALUE);
					dynAppend(Name,ERROR_VALUE);
					bError=true;
				}
				iNbLine++;
			} //End pattern type
			else //if(dsFileLines[l]!="")
			{
				if((bDefined)&(dsFileLines[l]==""))
				{
					iNbLine++;
					l++;
					if(patternMatch("*dpe=*", dsFileLines[l]))
					{
						sDpe=substr(dsFileLines[l], strpos(dsFileLines[l],"=")+1,strlen(dsFileLines[l]));
						l++;
						iNbLine++;
						if(patternMatch("*alias=*", dsFileLines[l]))
						{
							sAlias=substr(dsFileLines[l], strpos(dsFileLines[l],"=")+1,strlen(dsFileLines[l]));
							l++;			
							if(patternMatch("*name=*", dsFileLines[l]))
							{
								sName=substr(dsFileLines[l], strpos(dsFileLines[l],"=")+1,strlen(dsFileLines[l]));
								l++;						
								sSelect="X";
							} //End pattern name
							else
							{
								sSelect="F";
								sName=ERROR_VALUE;
								bError=true;
								if(!patternMatch("*range=*", dsFileLines[l])) l++;
							}			
						} //End pattern alias
						else
						{
							sSelect="F";
							sAlias=ERROR_VALUE;
							bError=true;
						}
						
					}// End pattern dpe
					else //dpe ERROR
					{
						sSelect="F";
						sDpe=ERROR_VALUE;
						sAlias=ERROR_VALUE;
						sName=ERROR_VALUE;
						bError=true;	
					}
                    
					dynAppend(Name,sName);
					dynAppend(Type,sType);
					dynAppend(Alias,sAlias);
					dynAppend(Dpe,sDpe);
                        
					if (sType == "[FLOAT]" || sType == "[INTEGER]" || sType == "[UNSIGNED]")	
                        iStart=1;
					else 
                        iStart=0;
						
					for(j=0; j<=len2-l; j++)
					{						
						iNbLine++;
						int temp = j+iStart;
						if(patternMatch("*range="+temp+"*", dsFileLines[j+l]))
						{
							dsSplitRange=strsplit(dsFileLines[j+l], ";");
							iLen_Range=dynlen(dsSplitRange);
									
							if (iLen_Range>=5)	
							{
                                //tha prepei na elenksw oti yparxoun ta range=;description=;ff=;fm=;fc=;
                                //meta an einai ta range 6 + na eleksw gia gia state
                                // an megalitera tou 7 + limit!
                                        
								if(j>0)
								{	
									dynAppend(Type,"");
									dynAppend(Alias,"");
									dynAppend(Dpe,"");
									dynAppend(Name,"");											
								}
                    
								dynAppend(Line,iNbLine);
								dynAppend(Select,sSelect);
								sDescription= substr(dsSplitRange[c_DESCRIPTION], strpos(dsSplitRange[c_DESCRIPTION],"=")+2,strlen(dsSplitRange[c_DESCRIPTION])-15);
								sFF= substr(dsSplitRange[c_FF], strpos(dsSplitRange[c_FF],"=")+1, strlen(dsSplitRange[c_FF]));
								sFM= substr(dsSplitRange[c_FM], strpos(dsSplitRange[c_FM],"=")+1, strlen(dsSplitRange[c_FM]));
								sFC= substr(dsSplitRange[c_FC], strpos(dsSplitRange[c_FC],"=")+1, strlen(dsSplitRange[c_FC]));
								if (iLen_Range>=c_STATE)
								{
									sState = substr(dsSplitRange[c_STATE], strpos(dsSplitRange[c_STATE],"=")+1,strlen(dsSplitRange[c_STATE]));
                                    if (iLen_Range>=c_Limit) //If state is specified check if any limit is specified
                                    {
                                        sLimit = substr(dsSplitRange[c_Limit], strpos(dsSplitRange[c_Limit],"limit")+5,strlen(dsSplitRange[c_Limit]));                      
                                    }
								}
								else
								{
									if(sType == "[FLOAT]" || sType == "[INTEGER]" || sType == "[UNSIGNED]") sState = "";
									if(sType == "[BOOLEAN]")  sState = "";	sLimit = "";
								}
								dynAppend(Range,temp);
								dynAppend(Description,sDescription);
								dynAppend(FF,sFF);
								dynAppend(FC,sFC);
								dynAppend(FM,sFM);                    
								dynAppend(State,sState);
								dynAppend(Limit,sLimit);                    
							} //end if iLen_Range >=5
							else
							{
								if(j>0)
								{	
									dynAppend(Type,"");
									dynAppend(Alias,"");
									dynAppend(Dpe,"");
									dynAppend(Name,"");
								}
								dynAppend(Description,ERROR_VALUE);
								dynAppend(Range,ERROR_VALUE);
								dynAppend(State,ERROR_VALUE);
								dynAppend(FF,ERROR_VALUE);
								dynAppend(FC,ERROR_VALUE);
								dynAppend(FM,ERROR_VALUE);
								dynAppend(Limit,ERROR_VALUE);                    
								bError=true;
							}
						}//End range j
						else
						{
							if(j>0)
							{	
								dynAppend(Type,"");
								dynAppend(Alias,"");
								dynAppend(Dpe,"");
								dynAppend(Name,"");
							}
							dynAppend(Line,iNbLine);
							dynAppend(Description,ERROR_VALUE);
							dynAppend(Range,ERROR_VALUE);
							dynAppend(State,ERROR_VALUE);
							dynAppend(FF,ERROR_VALUE);
							dynAppend(FC,ERROR_VALUE);
							dynAppend(FM,ERROR_VALUE);
							dynAppend(Limit,ERROR_VALUE);                                        
							dynAppend(Select,"F");
							bError=true;
						}//End range error
								
					}//End for
				}
				else //first element error
				{
					dynAppend(Select,"F");
					dynAppend(Type,ERROR_VALUE);
					dynAppend(Alias,ERROR_VALUE);
					dynAppend(Dpe,ERROR_VALUE);
					dynAppend(Line,iNbLine);		
					dynAppend(Description,ERROR_VALUE);
					dynAppend(Range,ERROR_VALUE);
					dynAppend(State,ERROR_VALUE);
					dynAppend(FF,ERROR_VALUE);
					dynAppend(FC,ERROR_VALUE);
					dynAppend(FM,ERROR_VALUE);
					dynAppend(Limit,ERROR_VALUE);                              
					dynAppend(Name,ERROR_VALUE);
					bError=true;
				}
			} //End else pattern type
		} //End for
	}//End if function defined
	else
	{
	    DebugN("DLL not defined");
	    bError=true;
	}//End function not defined
    
    progressBar.visible(false);
    
    mResult["Selected"] = Select;
    mResult["Type"] = Type;
    mResult["Alias"] = Alias;
    mResult["Dpe"] = Dpe;
    mResult["Name"] = Name;
    mResult["Description"] = Description;
    mResult["Range"] = Range;
    mResult["State"] = State;
    mResult["FF"] = FF;
    mResult["FM"] = FM;
    mResult["FC"] = FC;
    mResult["Limit"] = Limit;
    mResult["Line"] = Line;  
    mResult["bError"] = bError;
    return mResult;    
}


/** Convert .txt configuration file to .xml configuration file

@param xmlFilePath,   string, input, Filepath
@param m,             mapping, config data

@return               string, the new converted xml filepath.
*/
private string unLaser_convertToXML(string xmlFilePath, mapping &m)
{ 
    if(getExt(xmlFilePath) !="")
        xmlFilePath = delExt(xmlFilePath);
    xmlFilePath+=".xml";
  
    file xmlFile = fopen(xmlFilePath,"w"); 
    
    //root xml file node.
    string xmlnsXsi = unLaser_createXmlAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance"); 
    string xmlnsXSD = unLaser_createXmlAttribute("xmlns:laserXSD", "http://en-dep.web.cern.ch/en-dep/structure/ICE/"); 
    fputs("<laserXSD:configFile " + xmlnsXsi + xmlnsXSD + ">\n", xmlFile);       

    for(int i=1; i<=dynlen(m["Type"]); i++)
    {
        if(m["Selected"][i] == "F")
          continue;

        if(patternMatch("*[*]*", m["Type"][i]))
        {     
            string xmlDpe = unLaser_createXmlAttribute("Dpe", m["Dpe"][i]);
            string tempType = strrtrim(strltrim(m["Type"][i],"["),"]"); 
            string xmlType = unLaser_createXmlAttribute("Type", tempType);           
            string xmlAlias = unLaser_createXmlAttribute("Alias", m["Alias"][i]);     
            string xmlName = unLaser_createXmlAttribute("Name", m["Name"][i]);           

            fputs(" <Datapoint " + xmlDpe + xmlType + xmlAlias + xmlName + "T_mode=\"false\">\n", xmlFile); 
                  
            int j = i;
            do
            {
                if(m["Selected"][j] == "X") {
       
                    string xmlRange = unLaser_createXmlAttribute("Num", m["Range"][j]); 
                    strreplace(m["Description"][j], "\"", "");
                    string xmlDescription = unLaser_createXmlAttribute("Description", m["Description"][j]);      
                    string xmlFF = unLaser_createXmlAttribute("FF", m["FF"][j]);     
                    string xmlFM = unLaser_createXmlAttribute("FM", m["FM"][j]);       
                    string xmlFC = unLaser_createXmlAttribute("FC", m["FC"][j]);       
                    string xmlState = unLaser_createXmlAttribute("State", m["State"][j]);    
                    string xmlLimit = unLaser_createXmlAttribute("Limit", m["Limit"][j]);                        

                    fputs("  <Range " + xmlRange + xmlDescription + xmlFF + xmlFM + xmlFC + xmlState + xmlLimit + "/>\n", xmlFile);      
                }            
                    
                if (j == dynlen(m["Type"])) // if last element dont continue
                    break;
                                
                j++; 
                
            } while(m["Type"][j]=="");
            
            fputs(" </Datapoint>\n",xmlFile); 
        }
  }
  fputs("</laserXSD:configFile>\n",xmlFile); 
  fflush(xmlFile);  
  fclose(xmlFile); // close file
  return xmlFilePath;
}


private string unLaser_createXmlAttribute(string key,  string value)
{
    if(value != "") {
        return (key + "=\"" + value + "\" ");   
    }
    return ("");
}

